# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-10-30 19:33
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('jacc', '0002_auto_20171020_1331'),
    ]

    operations = [
        migrations.RenameField(
            model_name='accountentry',
            old_name='source',
            new_name='source_file',
        ),
    ]
