# Generated by Django 2.1.2 on 2018-10-08 23:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('jacc', '0017_auto_20180807_1808'),
    ]

    operations = [
        migrations.AlterField(
            model_name='invoice',
            name='number',
            field=models.CharField(blank=True, db_index=True, default='', max_length=32, verbose_name='invoice number'),
        ),
    ]
