# Generated by Django 2.0 on 2018-01-21 17:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('jacc', '0007_invoice_number'),
    ]

    operations = [
        migrations.RenameField(
            model_name='accounttype',
            old_name='is_debit',
            new_name='is_asset',
        ),
        migrations.AlterField(
            model_name='accounttype',
            name='is_asset',
            field=models.BooleanField(verbose_name='is asset'),
        ),
    ]
